/*
 * OpenR2 
 * MFC/R2 call setup library
 *
 * Moises Silva <moises.silva@gmail.com>
 * Copyright (C) Moises Silva
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include <zaptel/zaptel.h>
#include "r2chan.h"
#include "r2log.h"
#include "r2proto.h"
#include "r2utils.h"

#define OR2_PROTO_ANSWER_WAIT_TIME 150

#define GA_TONE(r2chan) (r2chan)->r2context->mf_ga_tones
#define GB_TONE(r2chan) (r2chan)->r2context->mf_gb_tones
#define GC_TONE(r2chan) (r2chan)->r2context->mf_gc_tones

#define GI_TONE(r2chan) (r2chan)->r2context->mf_g1_tones
#define GII_TONE(r2chan) (r2chan)->r2context->mf_g2_tones

#define TIMER(r2chan) (r2chan)->r2context->timers

/* Note that we compare >= because even if max_dnis is zero
   we could get 1 digit, want it or not :-) */
#define DNIS_COMPLETE(r2chan) ((r2chan)->dnis_len >= (r2chan)->r2context->max_dnis)


static void r2config_mexico(openr2_context_t *r2context)
{
	OR2_CONTEXT_STACK;

	/* Telmex, Avantel and most telcos in Mexico send DNIS first and ANI at the end, however
	   this can be modified by the user because I know of at least 1 telco (Maxcom)
	   which requires the ANI first and the DNIS later */
	r2context->get_ani_first = 0;

	r2context->timers.mf_back_cycle = 5000;

	/* Mexico use a special signal to request 
	   calling party category AND switch to Group C */
	r2context->mf_ga_tones.request_category = OR2_MF_TONE_INVALID;
	r2context->mf_ga_tones.request_category_and_change_to_gc = OR2_MF_TONE_6;

	/* GA next ANI is replaces by GC next ANI signal */
	r2context->mf_ga_tones.request_next_ani_digit = OR2_MF_TONE_INVALID;

	/* Group B */
	r2context->mf_gb_tones.accept_call_with_charge = OR2_MF_TONE_1;
	r2context->mf_gb_tones.accept_call_no_charge = OR2_MF_TONE_5;
	r2context->mf_gb_tones.busy_number = OR2_MF_TONE_2;
	r2context->mf_gb_tones.unassigned_number = OR2_MF_TONE_2;

	/* GROUP C */
	r2context->mf_gc_tones.request_next_ani_digit = OR2_MF_TONE_1;
	r2context->mf_gc_tones.request_change_to_g2 = OR2_MF_TONE_3;
	r2context->mf_gc_tones.request_next_dnis_digit_and_change_to_ga = OR2_MF_TONE_5;
	
	/* Mexico has no signal when running out of DNIS, 
	   timeout is used instead*/
	r2context->mf_g1_tones.no_more_dnis_available = OR2_MF_TONE_INVALID;
}

static const openr2_abcd_signal_t standard_abcd_signals[OR2_NUM_ABCD_SIGNALS] =
{
	/* OR2_ABCD_IDLE */ 0x8,
	/* OR2_ABCD_BLOCK */ 0xC,
	/* OR2_ABCD_SEIZE */ 0x0,
	/* OR2_ABCD_SEIZE_ACK */ 0xC,
	/* OR2_ABCD_CLEAR_BACK */ 0xC,
	/* OR2_ABCD_CLEAR_FORWARD */ 0x8,
	/* OR2_ABCD_ANSWER */ 0x4
};

static const char *abcd_names[OR2_NUM_ABCD_SIGNALS] =
{
	/* OR2_ABCD_IDLE */ "IDLE",
	/* OR2_ABCD_BLOCK */ "BLOCK",
	/* OR2_ABCD_SEIZE */ "SEIZE",
	/* OR2_ABCD_SEIZE_ACK */ "SEIZE_ACK",
	/* OR2_ABCD_CLEAR_BACK */ "CLEAR_BACK",
	/* OR2_ABCD_CLEAR_FORWARD */ "CLEAR_FORWARD",
	/* OR2_ABCD_ANSWER */ "ANSWER" 
};

/* this array must contain all the calling party category
   values specified in openr2_calling_party_categories_t
   except the UNKNOWN. Also must match the order in 
   openr2_mf_g2_tones_t */
static const openr2_calling_party_category_t r2_calling_categories[] =
{
	OR2_CALLING_PARTY_CATEGORY_NATIONAL_SUBSCRIBER,
	OR2_CALLING_PARTY_CATEGORY_NATIONAL_PRIORITY_SUBSCRIBER,
	OR2_CALLING_PARTY_CATEGORY_NATIONAL_MAINTENANCE,
	OR2_CALLING_PARTY_CATEGORY_NATIONAL_OPERATOR,
	OR2_CALLING_PARTY_CATEGORY_NATIONAL_DATA,
	OR2_CALLING_PARTY_CATEGORY_NATIONAL_PAYPHONE,

	OR2_CALLING_PARTY_CATEGORY_INTERNATIONAL_SUBSCRIBER,
	OR2_CALLING_PARTY_CATEGORY_INTERNATIONAL_PRIORITY_SUBSCRIBER,
	OR2_CALLING_PARTY_CATEGORY_INTERNATIONAL_OPERATOR,
	OR2_CALLING_PARTY_CATEGORY_INTERNATIONAL_DATA
};

static void (*r2variants[])(openr2_context_t *) =
{
	/* ALGERIA */ NULL,
	/* ARGENTINA */NULL,
	/* BAHRAIN */ NULL,
	/* BOLIVIA */ NULL,
	/* BRAZIL */ NULL,
	/* CHILE */ NULL,
	/* CHINA */ NULL,
	/* COLOMBIA_LAND */ NULL,
	/* COLOMBIA_CELL */ NULL,
	/* CZECH */ NULL,
	/* DEMO_CONGO */ NULL,
	/* EGYPT */ NULL,
	/* GHANA */ NULL,
	/* HONDURAS */ NULL,
	/* INDIA */ NULL,
	/* IRAQ */ NULL,
	/* ITU */ NULL,
	/* KOREA */ NULL,
	/* KUWAIT */ NULL,
	/* MALAYSIA */ NULL,
	/* MEXICO */ r2config_mexico,
	/* NIGERIA */ NULL,
	/* PANAMA */ NULL,
	/* PHILIPINNES */ NULL,
	/* ROMANIA */ NULL,
	/* SAUDI_ARABIA */ NULL,
	/* SINGAPORE */ NULL,
	/* THAILAND */ NULL,
	/* VENEZUELA */ NULL,
	/* VIETNAM */ NULL
};

static char *r2variants_names[] =
{
	/* ALGERIA */ NULL,
	/* ARGENTINA */NULL,
	/* BAHRAIN */ NULL,
	/* BOLIVIA */ NULL,
	/* BRAZIL */ NULL,
	/* CHILE */ NULL,
	/* CHINA */ NULL,
	/* COLOMBIA_LAND */ NULL,
	/* COLOMBIA_CELL */ NULL,
	/* CZECH */ NULL,
	/* DEMO_CONGO */ NULL,
	/* EGYPT */ NULL,
	/* GHANA */ NULL,
	/* HONDURAS */ NULL,
	/* INDIA */ NULL,
	/* IRAQ */ NULL,
	/* ITU */ NULL,
	/* KOREA */ NULL,
	/* KUWAIT */ NULL,
	/* MALAYSIA */ NULL,
	"MX" /* MEXICO */,
	/* NIGERIA */ NULL,
	/* PANAMA */ NULL,
	/* PHILIPINNES */ NULL,
	/* ROMANIA */ NULL,
	/* SAUDI_ARABIA */ NULL,
	/* SINGAPORE */ NULL,
	/* THAILAND */ NULL,
	/* VENEZUELA */ NULL,
	/* VIETNAM */ NULL
};

static int set_abcd_signal(openr2_chan_t *r2chan, openr2_abcd_signal_t signal)
{
	OR2_CHAN_STACK;
	int res, abcd;
	abcd = r2chan->r2context->abcd_signals[signal];
	openr2_log(r2chan, OR2_LOG_CAS_TRACE, "ABCD Tx >> [%s] 0x%X\n", abcd_names[signal], abcd);
	r2chan->abcd_write = abcd;
	/* set the NON R2 bits to 1 */
	abcd |= r2chan->r2context->abcd_nonr2_bits; 
	res = ioctl(r2chan->fd, ZT_SETTXBITS, &abcd);
	if (res) {
		EMI(r2chan)->on_os_error(r2chan, errno);
		return -1;
	} 
	return 0;
}

/* here we configure R2 as ITU and finally call a country specific function to alter the protocol description acording
   to the specified R2 variant */
int openr2_proto_configure_context(openr2_context_t *r2context, openr2_variant_t variant, int max_ani, int max_dnis)
{
	OR2_CONTEXT_STACK;

	/* if we don't know that variant, return failure */
	if (variant < 0 || variant > sizeof(r2variants)) {
		return -1;
	}

	/* if we have the variant, but we don't have a configuration 
	   function for it return failure */
	if (!r2variants[variant]) {
		return -1;
	}

	/* set default standard ABCD signals */
	memcpy(r2context->abcd_signals, standard_abcd_signals, sizeof(standard_abcd_signals));

	/* Default Non-R2 bit required to be on is D */
	r2context->abcd_nonr2_bits = 0x1;    /* 0001 */

	/* Default R2 bits are A and B */
	r2context->abcd_r2_bits = 0xC; /*  1100 */

	/* set default values for the protocol timers */
	r2context->timers.mf_back_cycle = 15000;
	r2context->timers.mf_back_resume_cycle = 150;
	r2context->timers.mf_fwd_safety = 30000;
	r2context->timers.r2_seize = 2000;
	r2context->timers.r2_answer = 60000; 

	/* Max ANI and DNIS */
	r2context->max_dnis = max_dnis;
	r2context->max_ani = max_ani;

	/* the forward R2 side always send DNIS first but
	   most variants continue by asking ANI first
	   and continuing with DNIS at the end  */
	r2context->get_ani_first = 1;

	/* Group A tones. Requests of ANI, DNIS and Calling Party Category */
	r2context->mf_ga_tones.request_next_dnis_digit = OR2_MF_TONE_1;
	r2context->mf_ga_tones.request_next_ani_digit = OR2_MF_TONE_5;
	r2context->mf_ga_tones.request_category = OR2_MF_TONE_5;
	r2context->mf_ga_tones.request_category_and_change_to_gc = OR2_MF_TONE_INVALID;
	r2context->mf_ga_tones.request_change_to_g2 = OR2_MF_TONE_3;

	/* Groub B tones. Decisions about what to do with the call */
	r2context->mf_gb_tones.accept_call_with_charge = OR2_MF_TONE_6;
	r2context->mf_gb_tones.accept_call_no_charge = OR2_MF_TONE_7;
	r2context->mf_gb_tones.busy_number = OR2_MF_TONE_3;
	r2context->mf_gb_tones.network_congestion = OR2_MF_TONE_4;
	r2context->mf_gb_tones.unassigned_number = OR2_MF_TONE_5;
	r2context->mf_gb_tones.line_out_of_order = OR2_MF_TONE_8;

	/* Group C tones. Similar to Group A but for Mexico */
	r2context->mf_gc_tones.request_next_ani_digit = OR2_MF_TONE_INVALID;
	r2context->mf_gc_tones.request_change_to_g2 = OR2_MF_TONE_INVALID;
	r2context->mf_gc_tones.request_next_dnis_digit_and_change_to_ga = OR2_MF_TONE_INVALID;

	/* Group I tones. Attend requests of Group A  */
	r2context->mf_g1_tones.no_more_dnis_available = OR2_MF_TONE_15;
	r2context->mf_g1_tones.no_more_ani_available = OR2_MF_TONE_15;
	r2context->mf_g1_tones.caller_ani_is_restricted = OR2_MF_TONE_INVALID;

	/* Group II tones. */
	r2context->mf_g2_tones.national_subscriber = OR2_MF_TONE_1;
	r2context->mf_g2_tones.national_priority_subscriber = OR2_MF_TONE_2;
	r2context->mf_g2_tones.national_maintenance = OR2_MF_TONE_3;
	r2context->mf_g2_tones.national_operator = OR2_MF_TONE_5;
	r2context->mf_g2_tones.national_data = OR2_MF_TONE_6;
	r2context->mf_g2_tones.national_payphone = OR2_MF_TONE_INVALID;
	r2context->mf_g2_tones.international_subscriber = OR2_MF_TONE_7;
	r2context->mf_g2_tones.international_data = OR2_MF_TONE_8;
	r2context->mf_g2_tones.international_priority_subscriber = OR2_MF_TONE_9;
	r2context->mf_g2_tones.international_operator = OR2_MF_TONE_10;

	/* now configure the country specific variations */
	r2variants[variant](r2context);
	return 0;
}

static const char *r2state2str(openr2_abcd_state_t r2state)
{
	switch ( r2state ) {
	case OR2_IDLE:
		return "Idle";
	case OR2_SEIZE_ACK_TXD:
		return "Seize ACK Transmitted";
	case OR2_ANSWER_TXD:
		return "Answer Transmitted";
	case OR2_CLEAR_BACK_TXD:
		return "Clear Back Transmitted";
	case OR2_CLEAR_BACK_TONE_RXD:
		return "Clear Back Tone Received";
	case OR2_CLEAR_FWD_TXD:
		return "Clear Forward Transmitted";
	case OR2_SEIZE_TXD:
		return "Seize Transmitted";
	case OR2_SEIZE_ACK_RXD:
		return "Seize ACK Received";
	case OR2_ACCEPT_RXD:
		return "Accept Received";
	case OR2_ANSWER_RXD:
		return "Answer Received";
	case OR2_ANSWER_RXD_MF_PENDING:
		return "Answer Received with MF Pending";
	default: 
		return "*Unknown*";
	}
}

static const char *mfstate2str(openr2_mf_state_t mf_state)
{
	switch ( mf_state ) {
	case OR2_MF_OFF_STATE:
		return "MF Engine Off";

	case OR2_MF_SEIZE_ACK_TXD:
		return "Seize ACK Transmitted";
	case OR2_MF_CATEGORY_RQ_TXD:
		return "Category Request Transmitted";
	case OR2_MF_DNIS_RQ_TXD:
		return "DNIS Request Transmitted";
	case OR2_MF_ANI_RQ_TXD:
		return "ANI Request Transmitted";
	case OR2_MF_CHG_GII_TXD:
		return "Change To Group II Request Transmitted";
	case OR2_MF_ACCEPTED_TXD:
		return "Accepted Call Transmitted";
	case OR2_MF_DISCONNECT_TXD:
		return "Disconnect Tone Transmitted";

	case OR2_MF_CATEGORY_TXD:
		return "Category Transmitted";
	case OR2_MF_DNIS_TXD:
		return "DNIS Digit Transmitted";
	case OR2_MF_DNIS_END_TXD:
		return "End of DNIS Transmitted";
	case OR2_MF_ANI_TXD:
		return "ANI Digit Transmitted";
	case OR2_MF_ANI_END_TXD:
		return "End of ANI Transmitted";
	case OR2_MF_WAITING_TIMEOUT:
		return "Waiting Far End Timeout";

	default:
		return "*Unknown*";
	}
}

const char *openr2_proto_get_error(openr2_protocol_error_t error)
{
	switch ( error ) {
	case OR2_INVALID_CAS_BITS:
		return "Invalid CAS";
	case OR2_INVALID_MF_TONE:
		return "Invalid Multi Frequency Tone";
	case OR2_BACK_MF_TIMEOUT:
		return "Multi Frequency Cycle Timeout";
	case OR2_SEIZE_TIMEOUT:
		return "Seize Timeout";
	case OR2_ANSWER_TIMEOUT:
		return "Answer Timeout";
	case OR2_INVALID_R2_STATE:
		return "Invalid R2 state";
	case OR2_INVALID_MF_STATE:
		return "Invalid Multy Frequency State";
	case OR2_INVALID_MF_GROUP:
		return "Invalid R2 Group";
	case OR2_FWD_SAFETY_TIMEOUT:
		return "Forward Safety Timeout";
	case OR2_BROKEN_MF_SEQUENCE:
		return "Broken MF Sequence";
	case OR2_LIBRARY_BUG:
		return "OpenR2 Library BUG";
	default:
		return "*Unknown*";
	}
}

static const char *mfgroup2str(openr2_mf_group_t mf_group)
{
	switch ( mf_group ) {
	case OR2_MF_NO_GROUP:
		return "No Group";

	case OR2_MF_BACK_INIT:
		return "Backward MF init";
	case OR2_MF_GA:
		return "Backward Group A";
	case OR2_MF_GB:
		return "Backward Group B";
	case OR2_MF_GC:
		return "Backward Group C";

	case OR2_MF_FWD_INIT:
		return "Forward MF init";
	case OR2_MF_GI:
		return "Forward Group I";
	case OR2_MF_GII:
		return "Forward Group II";
	case OR2_MF_GIII:
		return "Forward Group III";

	default:
		return "*Unknown*";
	}
}

const char *openr2_proto_get_disconnect_string(openr2_call_disconnect_reason_t reason)
{
	switch (reason) {
	case OR2_CAUSE_BUSY_NUMBER:
		return "Busy Number";
	case OR2_CAUSE_NETWORK_CONGESTION:
		return "Network Congestion";
	case OR2_CAUSE_UNASSIGNED_NUMBER:
		return "Unassigned Number";
	case OR2_CAUSE_OUT_OF_ORDER:
		return "Line Out Of Order";
	case OR2_CAUSE_UNSPECIFIED:
		return "Not Specified";
	case OR2_CAUSE_NORMAL_CLEARING:
		return "Normal Clearing";
	case OR2_CAUSE_NO_ANSWER:
		return "No Answer";
	default:
		return "*Unknown*";
	}
}

int openr2_proto_set_idle(openr2_chan_t *r2chan)
{
	OR2_CHAN_STACK;
	int ret;
	/* cancel any event we could be waiting for */
	openr2_chan_cancel_timer(r2chan);

	/* initialize all the proto and call stuff */
	r2chan->ani[0] = '\0';
	r2chan->ani_len = 0;
	r2chan->ani_ptr = NULL;
	r2chan->dnis[0] = '\0';
	r2chan->dnis_len = 0;
	r2chan->dnis_ptr = NULL;
	r2chan->caller_ani_is_restricted = 0;
	r2chan->caller_category = OR2_CALLING_PARTY_CATEGORY_UNKNOWN;
	r2chan->r2_state = OR2_IDLE;
	r2chan->mf_state = OR2_MF_OFF_STATE;
	r2chan->mf_group = OR2_MF_NO_GROUP;
	r2chan->call_state = OR2_CALL_IDLE;
	r2chan->direction = OR2_DIR_STOPPED;
	r2chan->answered = 0;
	r2chan->mf_write_tone = 0;
	r2chan->mf_read_tone = 0;

	/* set ABCD to IDLE */
	ret = set_abcd_signal(r2chan, OR2_ABCD_IDLE);
	if (-1 == ret) {
		r2chan->r2context->last_error = OR2_LIBERR_CANNOT_SET_IDLE;
		openr2_log(r2chan, OR2_LOG_ERROR, "failed to set channel %d to IDLE state.\n");
		return -1;
	}
	return 0;
}

static void handle_protocol_error(openr2_chan_t *r2chan, openr2_protocol_error_t reason)
{
	OR2_CHAN_STACK;
	openr2_log(r2chan, OR2_LOG_ERROR, "Protocol error. Reason = %s, R2 State = %s, MF state = %s, MF Group = %s\n", 
			openr2_proto_get_error(reason), r2state2str(r2chan->r2_state), mfstate2str(r2chan->mf_state), mfgroup2str(r2chan->mf_group));
	openr2_log(r2chan, OR2_LOG_DEBUG, "DNIS = %s, ANI = %s, Last MF Signal = %c\n", r2chan->dnis, r2chan->ani, 
			r2chan->mf_read_tone ? r2chan->mf_read_tone : 0x20);
	/* mute anything we may have */
	MFI(r2chan)->mf_select_tone(r2chan->mf_write_handle, 0);
	openr2_proto_set_idle(r2chan);
	EMI(r2chan)->on_protocol_error(r2chan, reason);
}

static void handle_incoming_call(openr2_chan_t *r2chan)
{
	OR2_CHAN_STACK;
	r2chan->r2_state = OR2_SEIZE_ACK_TXD;
	r2chan->mf_state = OR2_MF_SEIZE_ACK_TXD;
	r2chan->mf_group = OR2_MF_BACK_INIT;
	r2chan->direction = OR2_DIR_BACKWARD;
	set_abcd_signal(r2chan, OR2_ABCD_SEIZE_ACK);
	/* now that we have sent the seize ack, we expect the first MF tone. 
	   let's init our MF engine */
	MFI(r2chan)->mf_write_init(r2chan->mf_write_handle, 0);
	MFI(r2chan)->mf_read_init(r2chan->mf_read_handle, 1);
	/* notify the user that a new call is starting to arrive */
	EMI(r2chan)->on_call_init(r2chan);
}

static void mf_fwd_safety_timeout_expired(openr2_chan_t *r2chan, void *data)
{
	OR2_CHAN_STACK;
	handle_protocol_error(r2chan, OR2_FWD_SAFETY_TIMEOUT);
}

static void prepare_mf_tone(openr2_chan_t *r2chan, int tone)
{
	OR2_CHAN_STACK;
	int flush_write = ZT_FLUSH_WRITE, ret;
	/* put silence only if we have a write tone */
	if (!tone && r2chan->mf_write_tone) {
		openr2_log(r2chan, OR2_LOG_MF_TRACE, "MF Tx >> %c [OFF]\n", r2chan->mf_write_tone);
		if (ioctl(r2chan->fd, ZT_FLUSH, &flush_write)) {
			EMI(r2chan)->on_os_error(r2chan, errno);
			return;
		}
	} 
	/* just choose the tone if the last chosen tone is different */
	if (r2chan->mf_write_tone != tone) {
		ret = MFI(r2chan)->mf_select_tone(r2chan->mf_write_handle, tone);
		if (-1 == ret) {
			/* this is not a protocol error, but there is nothing else we can do anyway */
			openr2_log(r2chan, OR2_LOG_ERROR, "BUG: failed to select MF tone\n");
			handle_protocol_error(r2chan, OR2_LIBRARY_BUG);
			return;
		}
		if (tone) {
			openr2_log(r2chan, OR2_LOG_MF_TRACE, "MF Tx >> %c [ON]\n", tone);
		}	
		r2chan->mf_write_tone = tone;
	}	
}

static void mf_send_dnis(openr2_chan_t *r2chan)
{
	/* TODO: Handle sending of previous DNIS digits */
	OR2_CHAN_STACK;
	/* if there are still some DNIS to send out */
	if ( *r2chan->dnis_ptr ) {
		openr2_log(r2chan, OR2_LOG_DEBUG, "Sending DNIS digit %c\n", *r2chan->dnis_ptr);
		r2chan->mf_state = OR2_MF_DNIS_TXD;
		prepare_mf_tone(r2chan, *r2chan->dnis_ptr);
		r2chan->dnis_ptr++;
	/* if no more DNIS, and there is a signal for it, use it */
	} else if ( GI_TONE(r2chan).no_more_dnis_available ) {
		openr2_log(r2chan, OR2_LOG_DEBUG, "Sending unavailable DNIS signal\n");
		r2chan->mf_state = OR2_MF_DNIS_END_TXD;
		prepare_mf_tone(r2chan, GI_TONE(r2chan).no_more_dnis_available);
	} else {
		openr2_log(r2chan, OR2_LOG_DEBUG, "No more DNIS. Doing nothing, waiting for timeout.\n");
		/* the callee should timeout to detect end of DNIS and
		   resume the MF signaling */
		r2chan->mf_state = OR2_MF_WAITING_TIMEOUT;
		/* even when we are waiting the other end to timeout we
		   cannot wait forever, put a timer to make sure of that */
		openr2_chan_set_timer(r2chan, TIMER(r2chan).mf_fwd_safety, mf_fwd_safety_timeout_expired, NULL);
	}
}

static void report_call_disconnection(openr2_chan_t *r2chan, openr2_call_disconnect_reason_t reason)
{
	openr2_log(r2chan, OR2_LOG_NOTICE, "Far end disconnected. Reason: %s\n", openr2_proto_get_disconnect_string(reason));
	r2chan->call_state = OR2_CALL_DISCONNECTED;
	EMI(r2chan)->on_call_disconnect(r2chan, reason);
}

static void report_call_end(openr2_chan_t *r2chan)
{
	openr2_log(r2chan, OR2_LOG_DEBUG, "Call ended\n");
	openr2_proto_set_idle(r2chan);
	EMI(r2chan)->on_call_end(r2chan);
}

int openr2_proto_handle_abcd_change(openr2_chan_t *r2chan)
{
	OR2_CHAN_STACK;
	int abcd, res;
	res = ioctl(r2chan->fd, ZT_GETRXBITS, &abcd);
	if (res) {
		EMI(r2chan)->on_os_error(r2chan, errno);
		return -1;
	}
	openr2_log(r2chan, OR2_LOG_CAS_TRACE, "ABCD Rx << 0x%X\n", abcd);
	/* pick up only the R2 bits */
	abcd &= r2chan->r2context->abcd_r2_bits;
	/* If the R2 bits are the same as the last time we read
	   just ignore them */
	if (r2chan->abcd_read == abcd) {
		openr2_log(r2chan, OR2_LOG_DEBUG, "No change in bits\n");
		return 0;
	} else {
		openr2_log(r2chan, OR2_LOG_DEBUG, "Bits changed from 0x%X to 0x%X\n", r2chan->abcd_read, abcd);
	}
	r2chan->abcd_read = abcd;
	/* ok, bits have changed, we need to know in which 
	   ABCD state we are to know what to do */
	switch (r2chan->r2_state) {
	case OR2_IDLE:
		if (abcd == r2chan->r2context->abcd_signals[OR2_ABCD_BLOCK]) {
			EMI(r2chan)->on_line_blocked(r2chan);
			return 0;
		}
		if (abcd == r2chan->r2context->abcd_signals[OR2_ABCD_IDLE]) {
			EMI(r2chan)->on_line_idle(r2chan);
			return 0;
		}
		if (abcd == r2chan->r2context->abcd_signals[OR2_ABCD_SEIZE]) {
			/* we are in IDLE and just received a seize request
			   lets handle this new call */
			handle_incoming_call(r2chan);
		} else {
			handle_protocol_error(r2chan, OR2_INVALID_CAS_BITS);
		}
		break;
	case OR2_SEIZE_ACK_TXD:
	case OR2_ANSWER_TXD:
		/* if call setup already started or the call is answered 
		   the only valid bit pattern is a clear forward, everything
		   else is protocol error */
		if (abcd == r2chan->r2context->abcd_signals[OR2_ABCD_CLEAR_FORWARD]) {
			r2chan->r2_state = OR2_CLEAR_FWD_RXD;
			report_call_disconnection(r2chan, OR2_CAUSE_NORMAL_CLEARING);
		} else {
			handle_protocol_error(r2chan, OR2_INVALID_CAS_BITS);
		}
		break;
	case OR2_SEIZE_TXD:
		/* if we transmitted a seize we expect the seize ACK */
		if (abcd == r2chan->r2context->abcd_signals[OR2_ABCD_SEIZE_ACK]) {
			/* When the other side send us the seize ack, MF tones
			   can start, we start transmitting DNIS */
			openr2_chan_cancel_timer(r2chan);
			r2chan->r2_state = OR2_SEIZE_ACK_RXD;
			r2chan->mf_group = OR2_MF_GI;
			MFI(r2chan)->mf_write_init(r2chan->mf_write_handle, 1);
			MFI(r2chan)->mf_read_init(r2chan->mf_read_handle, 0);
			mf_send_dnis(r2chan);
		} else {
			handle_protocol_error(r2chan, OR2_INVALID_CAS_BITS);
		}
		break;
	case OR2_CLEAR_BACK_TXD:
		if (abcd == r2chan->r2context->abcd_signals[OR2_ABCD_CLEAR_FORWARD]) {
			report_call_end(r2chan);
		} else {
			handle_protocol_error(r2chan, OR2_INVALID_CAS_BITS);
		}
		break;
	case OR2_ACCEPT_RXD:
		/* once we got MF ACCEPT tone, we expect the ABCD Answer */
		if (abcd == r2chan->r2context->abcd_signals[OR2_ABCD_ANSWER]) {
			openr2_chan_cancel_timer(r2chan);
			r2chan->r2_state = OR2_ANSWER_RXD;
			r2chan->call_state = OR2_CALL_ANSWERED;
			r2chan->mf_state = OR2_MF_OFF_STATE;
			r2chan->answered = 1;
			EMI(r2chan)->on_call_answered(r2chan);
		} else if (abcd == r2chan->r2context->abcd_signals[OR2_ABCD_CLEAR_BACK]) {
			report_call_disconnection(r2chan, OR2_CAUSE_NORMAL_CLEARING);
		} else {
			handle_protocol_error(r2chan, OR2_INVALID_CAS_BITS);
		}
		break;
	case OR2_SEIZE_ACK_RXD:
		if (abcd == r2chan->r2context->abcd_signals[OR2_ABCD_ANSWER]) {
			/* sometimes, since ABCD signaling is faster than MF detectors we
			   may receive the ANSWER signal before actually receiving the
			   MF tone that indicates the call has been accepted. We
			   cannot turn off the tone detector because the tone off
			   condition is still missing */
			r2chan->r2_state = OR2_ANSWER_RXD_MF_PENDING;
		} else if (abcd == r2chan->r2context->abcd_signals[OR2_ABCD_CLEAR_BACK]) {
			openr2_log(r2chan, OR2_LOG_ERROR, "Wah! clear back before answering, we should handle this.\n");
		} else {
			handle_protocol_error(r2chan, OR2_INVALID_CAS_BITS);
		}
		break;
	case OR2_ANSWER_RXD_MF_PENDING:
	case OR2_ANSWER_RXD:
		if (abcd == r2chan->r2context->abcd_signals[OR2_ABCD_CLEAR_BACK]) {
			report_call_disconnection(r2chan, OR2_CAUSE_NORMAL_CLEARING);
		} else {
			handle_protocol_error(r2chan, OR2_INVALID_CAS_BITS);
		}	
		break;
	case OR2_CLEAR_BACK_TONE_RXD:
		if (abcd == r2chan->r2context->abcd_signals[OR2_ABCD_IDLE]) {
			openr2_proto_set_idle(r2chan);
		} else {
			handle_protocol_error(r2chan, OR2_INVALID_CAS_BITS);
		}
		break;
	case OR2_CLEAR_FWD_TXD:
		if (abcd == r2chan->r2context->abcd_signals[OR2_ABCD_IDLE]) {
			/* if the call never was answered don't report call end */
			if (!r2chan->answered) {
				openr2_proto_set_idle(r2chan);
			} else {
				report_call_end(r2chan);
			}	
		} else {
			handle_protocol_error(r2chan, OR2_INVALID_CAS_BITS);
		}
		break;
	default:
		handle_protocol_error(r2chan, OR2_INVALID_R2_STATE);
	}
	return 0;
}

int openr2_proto_accept_call(openr2_chan_t *r2chan, openr2_call_accept_t accept)
{
	OR2_CHAN_STACK;
	openr2_mf_tone_t tone;
	if (OR2_CALL_OFFERED != r2chan->call_state) {
		openr2_log(r2chan, OR2_LOG_WARNING, "Cannot accept call if the call has not been offered!\n");
		return -1;
	}
	r2chan->mf_state = OR2_MF_ACCEPTED_TXD;
	tone = (accept == OR2_ACCEPT_WITH_CHARGE) 
	     ? GB_TONE(r2chan).accept_call_with_charge
	     : GB_TONE(r2chan).accept_call_no_charge;
	prepare_mf_tone(r2chan, tone);		
	return 0;
}

int openr2_proto_answer_call(openr2_chan_t *r2chan)
{
	/* TODO: The user could decide to answer even before
	   calling openr2_proto_accept_call and thus we should
	   handle OR2_CALL_OFFERED state here as well */
	OR2_CHAN_STACK;
	if (r2chan->call_state != OR2_CALL_ACCEPTED) {
		openr2_log(r2chan, OR2_LOG_WARNING, "Cannot answer call if the call is not accepted first\n");
		return -1;
	}
	r2chan->call_state = OR2_CALL_ANSWERED;
	r2chan->r2_state = OR2_ANSWER_TXD;
	set_abcd_signal(r2chan, OR2_ABCD_ANSWER);
	EMI(r2chan)->on_call_answered(r2chan);
	r2chan->answered = 1;
	return 0;
}

static void request_calling_party_category(openr2_chan_t *r2chan)
{
	OR2_CHAN_STACK;
	int tone = GA_TONE(r2chan).request_category ? GA_TONE(r2chan).request_category : GA_TONE(r2chan).request_category_and_change_to_gc;
	r2chan->mf_group = OR2_MF_GC;
	r2chan->mf_state = OR2_MF_CATEGORY_RQ_TXD;
	prepare_mf_tone(r2chan, tone);
}

static void set_silence(openr2_chan_t *r2chan)
{
	OR2_CHAN_STACK;
	prepare_mf_tone(r2chan, 0);
	r2chan->mf_write_tone = 0;
}

static void mf_back_resume_cycle(openr2_chan_t *r2chan, void *data)
{
	OR2_CHAN_STACK;
	set_silence(r2chan);
}

static void request_change_to_g2(openr2_chan_t *r2chan)
{
	OR2_CHAN_STACK;
	/* request to change to group 2 can come from either from Group C (only for Mexico)
	   or Group A (All the world, including Mexico) */
	int change_tone = (OR2_MF_GC == r2chan->mf_group) ? GC_TONE(r2chan).request_change_to_g2
		                                          : GA_TONE(r2chan).request_change_to_g2;
	r2chan->mf_group = OR2_MF_GB;
	r2chan->mf_state = OR2_MF_CHG_GII_TXD;
	prepare_mf_tone(r2chan, change_tone);
}

static void mf_back_cycle_timeout_expired(openr2_chan_t *r2chan, void *data)
{
	OR2_CHAN_STACK;
	if (OR2_MF_TONE_INVALID == GI_TONE(r2chan).no_more_dnis_available
	     && r2chan->mf_group == OR2_MF_GA
	     && r2chan->mf_state == OR2_MF_DNIS_RQ_TXD) {
		openr2_log(r2chan, OR2_LOG_DEBUG, "MF cycle timed out, no more DNIS\n");
		/* the other end has run out of DNIS digits and were in a R2 variant that
		   does not support 'No More DNIS available' signal (ain't that silly?), and
		   those R2 variants let the backward end to timeout and resume the MF dance,
		   that's why we timed out waiting for more DNIS. Let's resume the MF signaling
		   and ask the calling party category (if needed). Since they are now in a silent 
		   state we will not get a 'tone off' condition, hence we need a timeout to mute 
		   our tone */
		openr2_chan_set_timer(r2chan, TIMER(r2chan).mf_back_resume_cycle, mf_back_resume_cycle, NULL);
		if (!r2chan->r2context->get_ani_first) {
			/* we were not asked to get the ANI first, hence when this
		           timeout occurs we know for sure we have not retrieved ANI yet,
		           let's retrieve it now. */
			request_calling_party_category(r2chan);
		} else {
			/* ANI must have been retrieved already (before DNIS),
			   let's go directly to GII, the final stage */
			request_change_to_g2(r2chan);
		}
	} else {
		openr2_log(r2chan, OR2_LOG_WARNING, "MF back cycle timed out!\n");
		handle_protocol_error(r2chan, OR2_BACK_MF_TIMEOUT);
	}	
}

static void request_next_dnis_digit(openr2_chan_t *r2chan)
{
	OR2_CHAN_STACK;
	openr2_mf_tone_t request_tone = (OR2_MF_GC == r2chan->mf_group) 
		                      ? GC_TONE(r2chan).request_next_dnis_digit_and_change_to_ga
		                      : GA_TONE(r2chan).request_next_dnis_digit;
	r2chan->mf_group = OR2_MF_GA;
	r2chan->mf_state = OR2_MF_DNIS_RQ_TXD;
	prepare_mf_tone(r2chan, request_tone);
}

static void mf_receive_expected_dnis(openr2_chan_t *r2chan, int tone)
{
	OR2_CHAN_STACK;
	if (OR2_MF_TONE_10 <= tone && OR2_MF_TONE_9 >= tone) {
		openr2_log(r2chan, OR2_LOG_DEBUG, "Getting DNIS digit %c\n", tone);
		r2chan->dnis[r2chan->dnis_len++] = tone;
		r2chan->dnis[r2chan->dnis_len] = '\0';
		openr2_log(r2chan, OR2_LOG_DEBUG, "DNIS so far: %s, expected length: %d\n", r2chan->dnis, r2chan->r2context->max_dnis);
		if (DNIS_COMPLETE(r2chan)) {
			openr2_log(r2chan, OR2_LOG_DEBUG, "Done getting DNIS!\n");
			/* if this is the first and last DNIS digit we have or
			   we were not required to get the ANI first, request it now, 
			   otherwise is time to go to GII signals */
			if (1 == r2chan->dnis_len || !r2chan->r2context->get_ani_first) {
				request_calling_party_category(r2chan);
			} else {
				request_change_to_g2(r2chan);
			} 
		} else if (1 == r2chan->dnis_len && r2chan->r2context->get_ani_first) {
			request_calling_party_category(r2chan);
		} else {
			request_next_dnis_digit(r2chan);
		}
	} else if (GI_TONE(r2chan).no_more_dnis_available == tone) {
		/* not sure if we ever could get no more dnis as first DNIS tone
		   but let's handle it just in case */
		if (0 == r2chan->dnis_len || !r2chan->r2context->get_ani_first) {
			request_calling_party_category(r2chan);
		} else {
			request_change_to_g2(r2chan);
		}
	} else {
		/* we were supposed to handle DNIS, but the tone
		   is not in the range of valid digits */
		handle_protocol_error(r2chan, OR2_INVALID_MF_TONE);
	}
}

static void mf_set_calling_party_category(openr2_chan_t *r2chan, int tone)
{
	OR2_CHAN_STACK;
	/* iterate over the Group II tones struct in the R2 context until we find
	   a matching tone. If we don't find a matching tone, then set
	   the category as UNKNOWN. We are trusting both openr2_mf_g2_tones_t
	   and r2_calling_categories to have the same size */
	size_t num_tones = sizeof(openr2_mf_g2_tones_t) / sizeof(openr2_mf_tone_t);
	unsigned i = 0;
	openr2_mf_tone_t *r2chan_tones = (openr2_mf_tone_t *)&GII_TONE(r2chan);
	for ( ; i < num_tones; i++ ) {
		if ( tone == *(r2chan_tones + i) ) {
			r2chan->caller_category = r2_calling_categories[i];
			return;
		}
	}
	r2chan->caller_category = OR2_CALLING_PARTY_CATEGORY_UNKNOWN;
}

static void mf_receive_expected_ani(openr2_chan_t *r2chan, int tone)
{
	OR2_CHAN_STACK;
	int next_ani_request_tone = GC_TONE(r2chan).request_next_ani_digit ? 
		                    GC_TONE(r2chan).request_next_ani_digit : 
				    GA_TONE(r2chan).request_next_ani_digit;
	/* no tone, just request next ANI if needed, otherwise
	   switch to Group B/II  */
	if (!tone || (OR2_MF_TONE_10 <= tone && OR2_MF_TONE_9 >= tone)) {
		/* if we have a tone, save it */
		if (tone) {
			openr2_log(r2chan, OR2_LOG_DEBUG, "Getting ANI digit %c\n", tone);
			r2chan->ani[r2chan->ani_len++] = tone;
			r2chan->ani[r2chan->ani_len] = '\0';
			openr2_log(r2chan, OR2_LOG_DEBUG, "ANI so far: %s, expected length: %d\n", r2chan->ani, r2chan->r2context->max_ani);
		}
		/* if we don't have a tone, or the ANI len is not enough yet, 
		   ask for more ANI */
		if (!tone || r2chan->r2context->max_ani > r2chan->ani_len) {
			r2chan->mf_state = OR2_MF_ANI_RQ_TXD;
			prepare_mf_tone(r2chan, next_ani_request_tone);
		} else {
			openr2_log(r2chan, OR2_LOG_DEBUG, "Done getting ANI!\n");
			if (!r2chan->r2context->get_ani_first || DNIS_COMPLETE(r2chan)) {
				request_change_to_g2(r2chan);
			} else {
				request_next_dnis_digit(r2chan);
			}	
		}
	/* they notify us about no more ANI available or the ANI 
	   is restricted AKA private */
	} else if ( tone == GI_TONE(r2chan).no_more_ani_available ||
		    tone == GI_TONE(r2chan).caller_ani_is_restricted ) {
		openr2_log(r2chan, OR2_LOG_DEBUG, "Got end of ANI\n");
		if ( tone == GI_TONE(r2chan).caller_ani_is_restricted ) {
			openr2_log(r2chan, OR2_LOG_DEBUG, "ANI is restricted\n");
			r2chan->caller_ani_is_restricted = 1;
		}	
		if (!r2chan->r2context->get_ani_first || DNIS_COMPLETE(r2chan)) {
			request_change_to_g2(r2chan);
		} else {
			request_next_dnis_digit(r2chan);
		}	
	} else {
		handle_protocol_error(r2chan, OR2_INVALID_MF_TONE);
	}
}

static void handle_forward_mf_tone(openr2_chan_t *r2chan, int tone)
{
	OR2_CHAN_STACK;
	/* schedule a new timer that will handle the timeout for our tone signal */
	openr2_chan_set_timer(r2chan, TIMER(r2chan).mf_back_cycle, mf_back_cycle_timeout_expired, NULL);
	switch (r2chan->mf_group) {
	/* we just sent the seize ACK and we are starting with the MF dance */
	case OR2_MF_BACK_INIT:
		switch (r2chan->mf_state) {
		case OR2_MF_SEIZE_ACK_TXD:
			/* after sending the seize ack we expect DNIS  */
			mf_receive_expected_dnis(r2chan, tone);
			break;
		default:
			handle_protocol_error(r2chan, OR2_INVALID_MF_STATE);
			break;
		}
		break;
	/* We are now at Group A signals, requesting DNIS or ANI, 
	   depending on the protocol variant */
	case OR2_MF_GA:
		/* ok we're at GROUP A signals. Let's see what was 
		   the last thing we did */
		switch (r2chan->mf_state) {
		/* we requested more DNIS */
		case OR2_MF_DNIS_RQ_TXD:
			/* then we receive more DNIS :) */
			mf_receive_expected_dnis(r2chan, tone);
			break;
		/* we requested the calling party category */
		case OR2_MF_CATEGORY_RQ_TXD:
			break;
		/* we requested more ANI */
		case OR2_MF_ANI_RQ_TXD:
			break;
		/* hu? WTF we were doing?? */
		default:
			handle_protocol_error(r2chan, OR2_INVALID_MF_STATE);
			break;
		}
		break;
	case OR2_MF_GB:
		switch (r2chan->mf_state) {
		case OR2_MF_CHG_GII_TXD:
			/* we cannot do anything by ourselves. The user has
			   to decide what to do. Let's inform him/her that
			   a new call is ready to be accepted or rejected */
			r2chan->call_state = OR2_CALL_OFFERED;
			EMI(r2chan)->on_call_offered(r2chan, r2chan->ani, r2chan->dnis, r2chan->caller_category);
			break;
		default:
			handle_protocol_error(r2chan, OR2_INVALID_MF_STATE);
			break;
		}
		break;
	/* Viva Mexico Cabrones!, solo Mexico tiene Grupo C ;)  */
	/* Group C is only for Mexico */
	case OR2_MF_GC:
		/* at this point, we either sent a category request, 
		   usually preceding the ANI request or we already sent
		   an ANI request. Anything else, is protocol error */
		switch (r2chan->mf_state) {
		/* we requested the calling party category */
		case OR2_MF_CATEGORY_RQ_TXD:
			mf_set_calling_party_category(r2chan, tone);
			if (r2chan->r2context->max_ani > 0) {
				mf_receive_expected_ani(r2chan, 0);
			} else {
				/* switch to Group B/II, we're ready to answer! */
				request_change_to_g2(r2chan);
			}
			break;
		/* we requested more ANI */
		case OR2_MF_ANI_RQ_TXD:
			mf_receive_expected_ani(r2chan, tone);
			break;
		/* yikes, we have an error! */
		default:
			handle_protocol_error(r2chan, OR2_INVALID_MF_STATE);
			break;
		}
		break;
	default:
		handle_protocol_error(r2chan, OR2_INVALID_MF_GROUP);
		break;
	}
}

static void ready_to_answer(openr2_chan_t *r2chan, void *data)
{
	OR2_CHAN_STACK;
	openr2_chan_cancel_timer(r2chan);
	EMI(r2chan)->on_call_accepted(r2chan);
}

static void handle_forward_mf_silence(openr2_chan_t *r2chan)
{
	OR2_CHAN_STACK;
	/* we just got silence. Let's silence our side as well.
	   depending on the MF hooks implementation this may be
	   immediate or not */
	set_silence(r2chan);
	switch (r2chan->mf_group) {
	case OR2_MF_GA:
		/* no further action required. The other end should 
		   handle our previous request */
		break;
	case OR2_MF_GB:
		switch (r2chan->mf_state) {
		case OR2_MF_CHG_GII_TXD:
			/* no further action required */
			break;
		case OR2_MF_ACCEPTED_TXD:
			/* MF dance has ended. The call has not been answered
			   but the user must decide when to answer. We don't
			   notify immediately because we still don't stop our
			   tone. I tried waiting for ZT_IOMUX_WRITEEMPTY event
			   but even then, it seems sometimes the other end requires
			   a bit of time to detect our tone off condition. If we
			   notify the user of the call accepted and he tries to
			   answer, setting the ABCD bits before the other end
			   detects our tone off condition can lead to the other
			   end to never really detect our ABCD answer state or
			   consider it a protocol error */
			r2chan->mf_state = OR2_MF_OFF_STATE;
			r2chan->call_state = OR2_CALL_ACCEPTED;
			openr2_chan_set_timer(r2chan, OR2_PROTO_ANSWER_WAIT_TIME, ready_to_answer, NULL);
			break;
		case OR2_MF_DISCONNECT_TXD:	
			/* we did not accept the call and sent some disconnect tone 
			   (busy, network congestion etc). The other end will take care
			   of clearing up the call.  */
			openr2_chan_cancel_timer(r2chan);
			break;
		default:
			handle_protocol_error(r2chan, OR2_INVALID_MF_STATE);
			break;
		}
		break;
	case OR2_MF_GC:
		/* no further action required. The other end should 
		   handle our previous request */
		break;
	default:
		handle_protocol_error(r2chan, OR2_INVALID_MF_GROUP);
	}
}

static void handle_backward_mf_tone(openr2_chan_t *r2chan, int tone)
{
	OR2_CHAN_STACK;
	/* if we had a safety timer, clean it up */
	openr2_chan_cancel_timer(r2chan);

	/* we are the forward side, each time we receive a tone from the
	   backward side, we just mute our tone */
	set_silence(r2chan);
}

static void mf_send_category(openr2_chan_t *r2chan)
{
	OR2_CHAN_STACK;
	/* iterate over categories to find the index of the proper tone to send */
	openr2_mf_tone_t category_tone = OR2_MF_TONE_INVALID;
	size_t num_tones = sizeof(openr2_mf_g2_tones_t) / sizeof(openr2_mf_tone_t);
	openr2_mf_tone_t *r2chan_tones = (openr2_mf_tone_t *)&GII_TONE(r2chan);
	unsigned i = 0;
	for ( ; i < num_tones; i++ ) {
		if ( r2_calling_categories[i] == r2chan->caller_category ) {
			category_tone = *(r2chan_tones + i);
		}
	}
	if (OR2_MF_TONE_INVALID == category_tone) {
		openr2_log(r2chan, OR2_LOG_ERROR, "No MF tone for category %s, using national subscriber\n", openr2_proto_get_category_string(r2chan->caller_category));
		category_tone = GII_TONE(r2chan).national_subscriber;
	}
	r2chan->mf_state = OR2_MF_CATEGORY_TXD;
	openr2_log(r2chan, OR2_LOG_DEBUG, "Sending category request %s\n", openr2_proto_get_category_string(r2chan->caller_category));
	prepare_mf_tone(r2chan, category_tone);
}

static void mf_send_ani(openr2_chan_t *r2chan)
{
	/* TODO: Handle sending of previous ANI digits */
	OR2_CHAN_STACK;
	/* if the pointer to ANI is null, that means the caller ANI is restricted */
	if (NULL == r2chan->ani_ptr) {
		openr2_log(r2chan, OR2_LOG_DEBUG, "Sending Restricted ANI\n");
		r2chan->mf_state = OR2_MF_DNIS_END_TXD;
		prepare_mf_tone(r2chan, GI_TONE(r2chan).caller_ani_is_restricted);
	/* ok, ANI is not restricted, let's see 
	   if there are still some ANI to send out */
	} else if (*r2chan->ani_ptr) {
		openr2_log(r2chan, OR2_LOG_DEBUG, "Sending ANI digit %c\n", *r2chan->ani_ptr);
		r2chan->mf_state = OR2_MF_ANI_TXD;
		prepare_mf_tone(r2chan, *r2chan->ani_ptr);
		r2chan->ani_ptr++;
	/* if no more ANI, and there is a signal for it, use it */
	} else if (GI_TONE(r2chan).no_more_ani_available) {
		openr2_log(r2chan, OR2_LOG_DEBUG, "Sending more ANI unavailable\n");
		r2chan->mf_state = OR2_MF_DNIS_END_TXD;
		prepare_mf_tone(r2chan, GI_TONE(r2chan).no_more_ani_available);
	} else {
		openr2_log(r2chan, OR2_LOG_DEBUG, "No more ANI, expecting timeout from the other side\n");
		/* the callee should timeout to detect end of ANI and
		   resume the MF signaling */
		r2chan->mf_state = OR2_MF_WAITING_TIMEOUT;
		/* even when we are waiting the other end to timeout we
		   cannot wait forever, put a timer to make sure of that */
		openr2_chan_set_timer(r2chan, TIMER(r2chan).mf_fwd_safety, mf_fwd_safety_timeout_expired, NULL);
	}
}

static void handle_group_a_request(openr2_chan_t *r2chan, int tone)
{
	OR2_CHAN_STACK;
	openr2_mf_tone_t request_category_tone = GA_TONE(r2chan).request_category ?
						 GA_TONE(r2chan).request_category :
						 GA_TONE(r2chan).request_category_and_change_to_gc;
	if (tone == GA_TONE(r2chan).request_next_dnis_digit) {
		mf_send_dnis(r2chan);
	} else if (tone == GA_TONE(r2chan).request_next_ani_digit) {
		mf_send_ani(r2chan);
	} else if (tone == request_category_tone) {
		if (request_category_tone == GA_TONE(r2chan).request_category_and_change_to_gc) {
			r2chan->mf_group = OR2_MF_GIII;
		}	
		mf_send_category(r2chan);
	} else if (tone == GA_TONE(r2chan).request_change_to_g2) {
		r2chan->mf_group = OR2_MF_GII;
		mf_send_category(r2chan);
	} else {
		handle_protocol_error(r2chan, OR2_INVALID_MF_TONE);
	}
}

static void handle_group_c_request(openr2_chan_t *r2chan, int tone)
{
	OR2_CHAN_STACK;
	if (tone == GC_TONE(r2chan).request_next_ani_digit) {
		mf_send_ani(r2chan);
	} else if (tone == GC_TONE(r2chan).request_change_to_g2) {
		/* requesting change to Group II means we should
		   send the calling party category again  */
		r2chan->mf_group = OR2_MF_GII;
		mf_send_category(r2chan);
	} else if (tone == GC_TONE(r2chan).request_next_dnis_digit_and_change_to_ga) {
		r2chan->mf_group = OR2_MF_GI;
		mf_send_dnis(r2chan);
	} else {
		handle_protocol_error(r2chan, OR2_INVALID_MF_TONE);
	}
}

static void r2_answer_timeout_expired(openr2_chan_t *r2chan, void *data)
{
	OR2_CHAN_STACK;
	report_call_disconnection(r2chan, OR2_CAUSE_NO_ANSWER);
}

static void handle_group_b_request(openr2_chan_t *r2chan, int tone)
{
	OR2_CHAN_STACK;
	if (tone == GB_TONE(r2chan).accept_call_with_charge 
	    || tone == GB_TONE(r2chan).accept_call_no_charge) {
		if (r2chan->r2_state == OR2_ANSWER_RXD_MF_PENDING) {
			/* they answered before we even detected they accepted,
			   lets just call on_call_accepted and immediately
			   on_call_answered */

			/* first accepted */
			r2chan->r2_state = OR2_ACCEPT_RXD;
			EMI(r2chan)->on_call_accepted(r2chan);

			/* now answered */
			openr2_chan_cancel_timer(r2chan);
			r2chan->r2_state = OR2_ANSWER_RXD;
			r2chan->call_state = OR2_CALL_ANSWERED;
			r2chan->mf_state = OR2_MF_OFF_STATE;
			r2chan->answered = 1;
			EMI(r2chan)->on_call_answered(r2chan);
		} else {
			/* They have accepted the call. We do nothing but
			   wait for answer. */
			r2chan->r2_state = OR2_ACCEPT_RXD;
			openr2_chan_set_timer(r2chan, TIMER(r2chan).r2_answer, r2_answer_timeout_expired, NULL);
			EMI(r2chan)->on_call_accepted(r2chan);
		}	
	} else if (tone == GB_TONE(r2chan).busy_number){
		r2chan->r2_state = OR2_CLEAR_BACK_TONE_RXD;
		report_call_disconnection(r2chan, OR2_CAUSE_BUSY_NUMBER);
	} else if (tone == GB_TONE(r2chan).network_congestion) {
		r2chan->r2_state = OR2_CLEAR_BACK_TONE_RXD;
		report_call_disconnection(r2chan, OR2_CAUSE_NETWORK_CONGESTION);
	} else if (tone == GB_TONE(r2chan).unassigned_number) {
		r2chan->r2_state = OR2_CLEAR_BACK_TONE_RXD;
		report_call_disconnection(r2chan, OR2_CAUSE_UNASSIGNED_NUMBER);
	} else if (tone == GB_TONE(r2chan).line_out_of_order) {
		r2chan->r2_state = OR2_CLEAR_BACK_TONE_RXD;
		report_call_disconnection(r2chan, OR2_CAUSE_OUT_OF_ORDER);
	} else {
		handle_protocol_error(r2chan, OR2_INVALID_MF_TONE);
	}
}

static void handle_backward_mf_silence(openr2_chan_t *r2chan, int tone)
{
	OR2_CHAN_STACK;
	/* the backward side has muted its tone, it is time to take
	   action depending on the tone they sent */
	switch (r2chan->mf_group) {
	case OR2_MF_GI:
		handle_group_a_request(r2chan, tone);
		break;
	case OR2_MF_GII:
		handle_group_b_request(r2chan, tone);
		break;
	case OR2_MF_GIII:
		handle_group_c_request(r2chan, tone);
		break;
	default:
		handle_protocol_error(r2chan, OR2_INVALID_MF_GROUP);
		break;
	}
}

static int timediff(struct timeval *t1, struct timeval *t2)
{
	int secdiff = 0;
	int msdiff = 0;
	if (t1->tv_sec == t2->tv_sec) {
		return ((t1->tv_usec - t2->tv_usec)/1000);
	}
	secdiff = t1->tv_sec - t2->tv_sec;	
	secdiff--;
	msdiff  = (t1->tv_usec) ? (t1->tv_usec/1000)          : 0;
	msdiff += (t2->tv_usec) ? (1000 - (t2->tv_usec/1000)) : 0;
	return msdiff;
}

static int check_threshold(openr2_chan_t *r2chan, int tone)
{
	OR2_CHAN_STACK;
	int res = 0;
	int tone_threshold = 0;
	struct timeval currtime = {0, 0};
	if (r2chan->r2context->mf_threshold) {
		if (r2chan->mf_threshold_tone != tone) {
			res = gettimeofday(&r2chan->mf_threshold_time, NULL);
			if (-1 == res) {
				openr2_log(r2chan, OR2_LOG_ERROR, "gettimeofday failed when setting threshold time\n");
				return -1;
			}
			r2chan->mf_threshold_tone = tone;
		}
		res = gettimeofday(&currtime, NULL);
		if (-1 == res) {
			openr2_log(r2chan, OR2_LOG_ERROR, "gettimeofday failed when checking tone length\n");
			return -1;
		}
		tone_threshold = timediff(&currtime, &r2chan->mf_threshold_time);
		if (tone_threshold < r2chan->r2context->mf_threshold) {
			if (tone) {
				openr2_log(r2chan, OR2_LOG_DEBUG, "Tone %c ignored\n", tone);
			} else {
				openr2_log(r2chan, OR2_LOG_DEBUG, "Tone off ignored\n");
			}	
			return -1;
		}
	}
	return 0;
}

void openr2_proto_handle_mf_tone(openr2_chan_t *r2chan, int tone)
{
	OR2_CHAN_STACK;
	if (tone) {

		/* since we get a continuous tone, this tone might be one that we
		   already handled but since the other end has not detected our tone
		   has not stopped its own tone. If it is the same
		   just ignore it, we already handled it */
		if (r2chan->mf_read_tone == tone) {
			return;
		}

		/* safety check. Each rx tone should be muted before changing the tone
		   hence the read tone right now should be 0 */
		if (r2chan->mf_read_tone != 0) {
			openr2_log(r2chan, OR2_LOG_ERROR, "Broken MF sequence got %c but never got tone off for tone %c!\n", tone, r2chan->mf_read_tone);
			handle_protocol_error(r2chan, OR2_BROKEN_MF_SEQUENCE);
			return;
		}

		/* do threshold checking if enabled */
		if (check_threshold(r2chan, tone)) {
			return;
		}

		openr2_log(r2chan, OR2_LOG_MF_TRACE, "MF Rx << %c [ON]\n", tone);
		r2chan->mf_read_tone = tone;

		/* handle the incoming MF tone */
		if ( OR2_DIR_BACKWARD == r2chan->direction ) {
			handle_forward_mf_tone(r2chan, tone);
		} else if ( OR2_DIR_FORWARD == r2chan->direction ) {
			handle_backward_mf_tone(r2chan, tone);
		} else {
			openr2_log(r2chan, OR2_LOG_ERROR, "BUG: invalid direction of R2 channel\n");
			handle_protocol_error(r2chan, OR2_LIBRARY_BUG);
		}

	} else {

		/* If we already detected the silence condition, ignore this one */
		if (0 == r2chan->mf_read_tone) {
			return;
		}
		if (check_threshold(r2chan, 0)) {
			return;
		}
		/* handle the silence condition */
		openr2_log(r2chan, OR2_LOG_MF_TRACE, "MF Rx << %c [OFF]\n", r2chan->mf_read_tone);
		if (OR2_DIR_BACKWARD == r2chan->direction) {
			handle_forward_mf_silence(r2chan);
		} else if (OR2_DIR_FORWARD == r2chan->direction) {
			/* when we are in forward we take action when the other side
			   silence its tone, not when receiving the tone */
			handle_backward_mf_silence(r2chan, r2chan->mf_read_tone);
		} else {
			openr2_log(r2chan, OR2_LOG_ERROR, "BUG: invalid direction of R2 channel\n");
			handle_protocol_error(r2chan, OR2_LIBRARY_BUG);
		}
		r2chan->mf_read_tone = 0;

	}
}

static void seize_timeout_expired(openr2_chan_t *r2chan, void *data)
{
	OR2_CHAN_STACK;
	openr2_log(r2chan, OR2_LOG_WARNING, "Seize Timeout Expired!\n");
	handle_protocol_error(r2chan, OR2_SEIZE_TIMEOUT);
}

int openr2_proto_make_call(openr2_chan_t *r2chan, const char *ani, const char *dnis, openr2_calling_party_category_t category)
{
	OR2_CHAN_STACK;
	/* we can dial only if we're in IDLE */
	if (r2chan->call_state != OR2_CALL_IDLE) {
		return -1;
	}
	/* try to handle last minute changes if any. 
	   This will detect IDLE lines if the last time the user checked
	   it was in some other state */
	openr2_proto_handle_abcd_change(r2chan);
	if (r2chan->abcd_read != r2chan->r2context->abcd_signals[OR2_ABCD_IDLE]) {
		openr2_log(r2chan, OR2_LOG_ERROR, "Trying to dial out in a non-idle channel\n");
		return -1;
	}
	if (set_abcd_signal(r2chan, OR2_ABCD_SEIZE)) {
		return -1;
	}	
	r2chan->call_state = OR2_CALL_DIALING;
	r2chan->r2_state = OR2_SEIZE_TXD;
	r2chan->mf_group = OR2_MF_FWD_INIT;
	r2chan->direction = OR2_DIR_FORWARD;
	r2chan->caller_category = category;
	strncpy(r2chan->ani, ani, sizeof(r2chan->ani)-1);
	r2chan->ani[sizeof(r2chan->ani)-1] = '\0';
	r2chan->ani_ptr = r2chan->ani;
	strncpy(r2chan->dnis, dnis, sizeof(r2chan->dnis)-1);
	r2chan->dnis[sizeof(r2chan->dnis)-1] = '\0';
	r2chan->dnis_ptr = r2chan->dnis;
	openr2_chan_set_timer(r2chan, TIMER(r2chan).r2_seize, seize_timeout_expired, NULL);
	return 0;
}

static void send_disconnect(openr2_chan_t *r2chan, openr2_call_disconnect_reason_t reason)
{
	int tone;
	r2chan->mf_state = OR2_MF_DISCONNECT_TXD;
	switch (reason) {
	case OR2_CAUSE_BUSY_NUMBER:
		tone = GB_TONE(r2chan).busy_number;
		break;
	case OR2_CAUSE_NETWORK_CONGESTION:
		tone = GB_TONE(r2chan).network_congestion;
		break;
	case OR2_CAUSE_UNASSIGNED_NUMBER:
		tone = GB_TONE(r2chan).unassigned_number;
		break;
	case OR2_CAUSE_OUT_OF_ORDER:
		tone = GB_TONE(r2chan).line_out_of_order;
		break;
	default:
		tone = GB_TONE(r2chan).line_out_of_order;
		break;
	}
	prepare_mf_tone(r2chan, tone);
}

static void send_clear_forward(openr2_chan_t *r2chan)
{
	OR2_CHAN_STACK;
	r2chan->r2_state = OR2_CLEAR_FWD_TXD;
	r2chan->mf_state = OR2_MF_OFF_STATE;
	set_abcd_signal(r2chan, OR2_ABCD_CLEAR_FORWARD);
}

static void send_clear_backward(openr2_chan_t *r2chan)
{
	OR2_CHAN_STACK;
	r2chan->r2_state = OR2_CLEAR_BACK_TXD;
	r2chan->mf_state = OR2_MF_OFF_STATE;
	set_abcd_signal(r2chan, OR2_ABCD_CLEAR_BACK);	
}

int openr2_proto_disconnect_call(openr2_chan_t *r2chan, openr2_call_disconnect_reason_t reason)
{
	OR2_CHAN_STACK;
	/* cannot drop a call when there is none to drop */
	if (r2chan->call_state == OR2_CALL_IDLE) {
		return -1;
	}
	if (r2chan->direction == OR2_DIR_BACKWARD) {
		if (r2chan->call_state == OR2_CALL_OFFERED) {
			/* if the call has been offered we need to give a reason 
			   to disconnect using a MF tone. That should make the other
			   end send us a clear forward  */
			send_disconnect(r2chan, reason);
		} else if (r2chan->r2_state == OR2_CLEAR_FWD_RXD){
			/* if the user want to hangup the call and the other end
			   already said they want too, then just report the call end event */
			report_call_end(r2chan);
		} else {
			/* this is a normal clear backward */
			send_clear_backward(r2chan);
		}
	} else {
		send_clear_forward(r2chan);
	}
	return 0;
}

const char *openr2_proto_get_category_string(openr2_calling_party_category_t category)
{
	switch (category) {
	case OR2_CALLING_PARTY_CATEGORY_NATIONAL_SUBSCRIBER:
		return "National Subscriber";
	case OR2_CALLING_PARTY_CATEGORY_NATIONAL_PRIORITY_SUBSCRIBER:
		return "National Priority Subscriber";
	case OR2_CALLING_PARTY_CATEGORY_NATIONAL_MAINTENANCE:
		return "National Maintenance";
	case OR2_CALLING_PARTY_CATEGORY_NATIONAL_OPERATOR:
		return "National Operator";
	case OR2_CALLING_PARTY_CATEGORY_NATIONAL_DATA:
		return "National Data";
	case OR2_CALLING_PARTY_CATEGORY_NATIONAL_PAYPHONE:
		return "National PayPhone";

	case OR2_CALLING_PARTY_CATEGORY_INTERNATIONAL_SUBSCRIBER:
		return "International Subscriber";
	case OR2_CALLING_PARTY_CATEGORY_INTERNATIONAL_PRIORITY_SUBSCRIBER:
		return "International Priority Subscriber";
	case OR2_CALLING_PARTY_CATEGORY_INTERNATIONAL_OPERATOR:
		return "International Operator";
	case OR2_CALLING_PARTY_CATEGORY_INTERNATIONAL_DATA:
		return "International Data";

	default:
		return "*Unknown*";
	}
}

openr2_calling_party_category_t openr2_proto_get_category(const char *category)
{
	if (!strncasecmp(category, "NATIONAL_SUBSCRIBER", sizeof("NATIONAL_SUBSCRIBER")-1)) {
		return OR2_CALLING_PARTY_CATEGORY_NATIONAL_SUBSCRIBER;
	} else if (!strncasecmp(category, "NATIONAL_PRIORITY_SUBSCRIBER", sizeof("NATIONAL_PRIORITY_SUBSCRIBER")-1)) {
		return OR2_CALLING_PARTY_CATEGORY_NATIONAL_PRIORITY_SUBSCRIBER;
	} else if (!strncasecmp(category, "NATIONAL_MAINTENANCE", sizeof("NATIONAL_MAINTENANCE")-1)) {
		return OR2_CALLING_PARTY_CATEGORY_NATIONAL_MAINTENANCE;
	} else if (!strncasecmp(category, "NATIONAL_OPERATOR", sizeof("NATIONAL_OPERATOR")-1)) {
		return OR2_CALLING_PARTY_CATEGORY_NATIONAL_OPERATOR;
	} else if (!strncasecmp(category, "NATIONAL_DATA", sizeof("NATIONAL_DATA")-1)) {
		return OR2_CALLING_PARTY_CATEGORY_NATIONAL_DATA;
	} else if (!strncasecmp(category, "NATIONAL_PAYPHONE", sizeof("NATIONAL_PAYPHONE")-1)) {
		return OR2_CALLING_PARTY_CATEGORY_NATIONAL_PAYPHONE;
	} else if (!strncasecmp(category, "INTERNATIONAL_SUBSCRIBER", sizeof("INTERNATIONAL_SUBSCRIBER")-1)) {
		return OR2_CALLING_PARTY_CATEGORY_INTERNATIONAL_SUBSCRIBER;
	} else if (!strncasecmp(category, "INTERNATIONAL_PRIORITY_SUBSCRIBER", sizeof("INTERNATIONAL_PRIORITY_SUBSCRIBER")-1)) {
		return OR2_CALLING_PARTY_CATEGORY_INTERNATIONAL_PRIORITY_SUBSCRIBER;
	} else if (!strncasecmp(category, "INTERNATIONAL_OPERATOR", sizeof("INTERNATIONAL_OPERATOR")-1)) {
		return OR2_CALLING_PARTY_CATEGORY_INTERNATIONAL_OPERATOR;
	} else if (!strncasecmp(category, "INTERNATIONAL_DATA", sizeof("INTERNATIONAL_DATA")-1)) {
		return OR2_CALLING_PARTY_CATEGORY_INTERNATIONAL_DATA;
	}	
	return OR2_CALLING_PARTY_CATEGORY_UNKNOWN;
}

openr2_variant_t openr2_proto_get_variant(const char *variant_name)
{
	int i;

	for (i = 0; i < sizeof(r2variants_names)/sizeof(char *); i++) {
		if (r2variants_names[i] && !strncasecmp(r2variants_names[i], variant_name, sizeof(r2variants[i])-1)) {
			return (openr2_variant_t)i;
		}
	}
	return OR2VAR_UNKNOWN;
}

const char *openr2_proto_get_variant_string(openr2_variant_t variant)
{
	if (variant < 0 || variant > (sizeof(r2variants_names)/sizeof(char *))) {
		return "*Unknown*";
	}	
	return r2variants_names[variant];
}

const char *openr2_proto_get_state(openr2_chan_t *r2chan, int tx)
{
	OR2_CHAN_STACK;
	/* Since the state does not depend on ABCD bits only (some of them are repated)
	   we should fix this. Currently only works to detect IDLE,BLOCK conditions, but
	   even those could be false-detected */
	int i;
	openr2_abcd_signal_t signal = tx ? r2chan->abcd_write : r2chan->abcd_read;
	for (i = 0; i < sizeof(r2chan->r2context->abcd_signals)/sizeof(r2chan->r2context->abcd_signals[0]); i++) {
		if (r2chan->r2context->abcd_signals[i] == signal) {
			return abcd_names[i];
		}
		
	}
	return "*Unknown*";
}


